import matplotlib.pyplot as plt
import numpy as np


def simple_contour(
    fun,
    min_x=-1,
    max_x=1,
    min_y=-1,
    max_y=1,
    min_z=None,
    max_z=None,
    n_samples=100,
    n_levels=30,
):
    """
    Simple contour plot

    Args:
        fun (_type_): _description_
        min_x (int, optional): _description_. Defaults to -1.
        max_x (int, optional): _description_. Defaults to 1.
        min_y (int, optional): _description_. Defaults to -1.
        max_y (int, optional): _description_. Defaults to 1.
        min_z (int, optional): _description_. Defaults to 0.
        max_z (int, optional): _description_. Defaults to 1.
        n_samples (int, optional): _description_. Defaults to 100.
        n_levels (int, optional): _description_. Defaults to 5.

    Example:
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> from spotPython.fun.objectivefunctions import analytical
        >>> fun = analytical().fun_branin
        >>> simple_contour(fun=fun, n_levels=30, min_x=-5, max_x=10, min_y=0, max_y=15)

    """
    XX, YY = np.meshgrid(np.linspace(min_x, max_x, n_samples), np.linspace(min_y, max_y, n_samples))
    zz = np.array([fun(np.array([xi, yi]).reshape(-1, 2)) for xi, yi in zip(np.ravel(XX), np.ravel(YY))]).reshape(
        n_samples, n_samples
    )
    fig, ax = plt.subplots(figsize=(5, 2.7), layout="constrained")
    if min_z is None:
        min_z = np.min(zz)
    if max_z is None:
        max_z = np.max(zz)
    plt.contourf(
        XX,
        YY,
        zz,
        levels=np.linspace(min_z, max_z, n_levels),
        zorder=1,
        cmap="jet",
        vmin=min_z,
        vmax=max_z,
    )
    plt.colorbar()



def test_aggregate_mean_var_var():
    """
    Test aggregate mean and var.
    Combine a (6,3) array (X):
    [[1. 1. 1.]
    [1. 1. 1.]
    [2. 2. 2.]
    [2. 2. 2.]
    [1. 1. 1.]
    [1. 1. 1.]]
    and an (1,) array (y):
    [3. 3. 6. 6. 6. 6.]
    Expected results:
    array([[1., 1., 1.],
            [2., 2., 2.]]),
    array([4.5, 6. ]),
    array([3., 0.]).

    """
    import numpy as np
    from spotPython.utils.aggregate import aggregate_mean_var
    X_1 = np.ones((2,3))
    y_1 = np.sum(X_1, axis=1)
    y_2 = 2 * y_1
    X_2 = np.append(X_1, 2*X_1, axis=0)
    X = np.append(X_2, X_1, axis=0)
    y = np.append(y_1, y_2, axis=0)
    y = np.append(y, y_2, axis=0)
    Z = aggregate_mean_var(X,y)
    assert (Z[0] == np.array([[1., 1., 1.],[2., 2., 2.]])).all()
    assert (Z[1] == np.array([4.5, 6. ])).all()
    assert (Z[2] == np.array([3., 0.])).all()



from numpy import around
from numpy import isfinite


def repair_non_numeric(X, var_type):
    """
    Round non-numeric values to integers.

    Args:
        X (numpy.ndarray): X array
        var_type (list): list with type information
    """
    for i in range(X.shape[1]):
        if var_type[i] != "num":
            X[:, i] = around(X[:, i])
    return X


def remove_nan(X, y):
    ind = isfinite(y)
    y = y[ind]
    X = X[ind, :]
    return X, y



# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.0.6.dev0+g81f52a4.d20230202'
__version_tuple__ = version_tuple = (0, 0, 6, 'dev0', 'g81f52a4.d20230202')



from numpy.random import default_rng


class designs:
    """
    Super class for all design classes (factorial and spacefilling)
    """

    def __init__(self, k=2, seed=123):
        self.designs = []
        self.k = k
        self.seed = seed
        self.rng = default_rng(self.seed)

    def get_dim(self):
        """Return design dimension."""
        print(self.k)



from numpy.random import default_rng


class surrogates:
    """
    Super class for all surrogate model classes (e.g., Kriging)
    """
    def __init__(self, name="", seed=123, verbosity=0):
        self.name = name
        self.seed = seed
        self.rng = default_rng(self.seed)
        self.log = {}
        self.verbosity = verbosity



from .designs import designs
from scipy.stats.qmc import LatinHypercube
from numpy import repeat
from spotPython.utils.transform import scale
from numpy import zeros
from numpy import ones


class spacefilling(designs):
    def __init__(self, k=2, seed=123):
        """
        Spacefilling design class

        Args:
            k (int, optional): number of design variables (dimensions). Defaults to 2.
            seed (int, optional): random seed. Defaults to 123.
        """
        self.k = k
        self.seed = seed
        super().__init__(k, seed)
        self.sampler = LatinHypercube(d=self.k, seed=self.seed)

    def scipy_lhd(self, n, repeats=1, lower=None, upper=None):
        """
        Latin hypercube sampling based on scipy.

        Args:
            n (int): number of samples
            repeats (int): number of repeats (replicates)
            lower (int, optional): lower bound. Defaults to 0.
            upper (int, optional): upper bound. Defaults to 1.

        Returns:
            (numpy.ndarray): Latin hypercube design.
        """
        if lower is None:
            lower = zeros(self.k)
        if upper is None:
            upper = ones(self.k)
        sample = self.sampler.random(n=n)
        des = scale(sample, lower, upper)
        return repeat(des, repeats, axis=0)



def test_build_psi_vec():
    """
    Test build_psi_vec
    """
    import numpy as np
    from spotPython.spot import spot
    from spotPython.build.kriging import Kriging

    X_train = np.array([[1., 2.],
                        [2., 4.],
                        [3., 6.]])
    y_train = np.array([1., 2., 3.])

    S = Kriging(name='kriging',
                seed=123,
                log_level=50,
                n_theta=1,
                noise=False,
                cod_type="norm")
    S.fit(X_train, y_train)

    # force theta to simple values:
    S.theta = np.array([0.0])
    cod_x = np.array([1., 0.])
    S.psi = np.zeros((S.n, 1))
    S.build_psi_vec(cod_x)
    res = np.array([[np.exp(-1)],
        [np.exp(-1/2)],
        [np.exp(-1)]])

    assert np.array_equal(S.psi, res)



from sys import stdout


def progress_bar(progress, bar_length=10):
    """
    Displays or updates a console progress bar.
    See: https://stackoverflow.com/a/15860757

    Args:
        (float) progress: a float between 0 and 1. Any int will be converted to a float.
        A value under 0 represents a halt.
        A value at 1 or bigger represents 100%.
    """

    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(bar_length * progress))
    text = "spotPython tuning: [{0}] {1:.2f}% {2}\r".format(
        "#" * block + "-" * (bar_length - block), progress * 100, status
    )
    stdout.write(text)
    stdout.flush()



from numpy import mgrid
from .designs import designs


class factorial(designs):
    """
    Super class for factorial designs.
    """

    def __init__(self, k=2, seed=123):
        super().__init__(k, seed)

    def full_factorial(self, p):
        i = (slice(0, 1, p * 1j),) * self.k
        return mgrid[i].reshape(self.k, p**self.k).T



def test_infill():
    """
    Test infill method
    """
    import numpy as np
    from spotPython.fun.objectivefunctions import analytical
    from spotPython.spot.spot import Spot
    from spotPython.utils.repair import repair_non_numeric
    from sklearn import linear_model
    import pytest

    fun = analytical().fun_sphere

    spot_test = Spot(
        fun=fun,
        lower=np.array([-10,-1]),
        upper=np.array([10,1]),
    )


    # (S-2) Initial Design:
    spot_test.X = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                            repeats=spot_test.design_control["repeats"], 
                                            lower=spot_test.lower,
                                            upper=spot_test.upper)
    spot_test.X = repair_non_numeric(spot_test.X,
                                        spot_test.var_type)

    # (S-3): Eval initial design:
    spot_test.y = spot_test.fun(spot_test.X)
    spot_test.surrogate.fit(spot_test.X, spot_test.y)

    x = spot_test.X[0, :]
    y = spot_test.y[0]
    
    y0 = spot_test.infill(x)
    assert y0.shape[0] == 1
    assert y0.ndim == 1    
    # Kriging is interpolating, so y and yhat should be similar:
    assert y0 == pytest.approx(y, 0.1)
    
    # 2nd test with sklearn surrogate (a simple linear model):
    
    S_LM = linear_model.LinearRegression()
    spot_test = Spot(
        fun=fun,
        lower=np.array([-10,-1]),
        upper=np.array([10,1]),
        surrogate=S_LM
    )
    # (S-2) Initial Design:
    spot_test.X = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                            repeats=spot_test.design_control["repeats"], 
                                            lower=spot_test.lower,
                                            upper=spot_test.upper)
    spot_test.X = repair_non_numeric(spot_test.X,
                                        spot_test.var_type)

    # (S-3): Eval initial design:
    spot_test.y = spot_test.fun(spot_test.X)
    spot_test.surrogate.fit(spot_test.X, spot_test.y)

    x = spot_test.X[0, :]
    y0 = spot_test.infill(x)
    assert y0.shape[0] == 1
    assert y0.ndim == 1

    



def test_evaluate_new_X():
    """
    Test evaluation of suggest_new_X points suggested by the surrogate
    """
    import numpy as np
    from spotPython.spot import spot
    from spotPython.fun.objectivefunctions import analytical

    nn = 3
    fun_sphere = analytical().fun_sphere
    spot_1 = spot.Spot(
        fun=fun_sphere,
        lower=np.array([-1, -1]),
        upper=np.array([1, 1]),
        n_points=nn
    )

    # (S-2) Initial Design:
    spot_1.X = spot_1.design.scipy_lhd(
        spot_1.design_control["init_size"], lower=spot_1.lower, upper=spot_1.upper
    )
    print(spot_1.X)

    # (S-3): Eval initial design:
    spot_1.y = spot_1.fun(spot_1.X)
    print(spot_1.y)

    spot_1.surrogate.fit(spot_1.X, spot_1.y)
    X0 = spot_1.suggest_new_X()
    X0.size
    assert X0.size == spot_1.n_points * spot_1.k
    assert X0.ndim == 2
    assert X0.shape[0] == nn
    assert X0.shape[1] == 2



"""
OCBA: Optimal Computing Budget Allocation
"""

from spotPython.utils.aggregate import get_ranks
from numpy import int32, float64
from numpy import argpartition, repeat
from numpy import zeros, square, sqrt, full, around


def get_ocba(means, vars, delta):
    """
    Optimal Computer Budget Allocation (OCBA)

    References:
        Chun-Hung Chen and Loo Hay Lee:
        Stochastic Simulation Optimization: An Optimal Computer Budget Allocation,
        pp. 49 and pp. 215

        C.S.M Currie and T. Monks:
        How to choose the best setup for a system. A tutorial for the Simulation Workshop 2021,
        see:
        https://colab.research.google.com/github/TomMonks/sim-tools/blob/master/examples/sw21_tutorial.ipynb
        and
        https://github.com/TomMonks/sim-tools

    Examples:

        From the Chen et al. book (p. 49):
        mean_y = np.array([1,2,3,4,5])
        var_y = np.array([1,1,9,9,4])
        get_ocba(mean_y, var_y, 50)

        [11  9 19  9  2]

        Args:
        means (numpy.array):
            means
        vars (numpy.array):
            variances
        delta (int):
            incremental budget

        Returns:
        (numpy.array):
            budget recommendations. `(n,)` numpy.array
    """
    n_designs = means.shape[0]
    allocations = zeros(n_designs, int32)
    ratios = zeros(n_designs, float64)
    budget = delta
    ranks = get_ranks(means)
    best, second_best = argpartition(ranks, 2)[:2]
    ratios[second_best] = 1.0
    select = [i for i in range(n_designs) if i not in [best, second_best]]
    temp = (means[best] - means[second_best]) / (means[best] - means[select])
    ratios[select] = square(temp) * (vars[select] / vars[second_best])
    select = [i for i in range(n_designs) if i not in [best]]
    temp = (square(ratios[select]) / vars[select]).sum()
    ratios[best] = sqrt(vars[best] * temp)
    more_runs = full(n_designs, True, dtype=bool)
    add_budget = zeros(n_designs, dtype=float)
    more_alloc = True
    while more_alloc:
        more_alloc = False
        ratio_s = (more_runs * ratios).sum()
        add_budget[more_runs] = (budget / ratio_s) * ratios[more_runs]
        add_budget = around(add_budget).astype(int)
        mask = add_budget < allocations
        add_budget[mask] = allocations[mask]
        more_runs[mask] = 0
        if mask.sum() > 0:
            more_alloc = True
        if more_alloc:
            budget = allocations.sum() + delta
            budget -= (add_budget * ~more_runs).sum()
    t_budget = add_budget.sum()
    add_budget[best] += allocations.sum() + delta - t_budget
    return add_budget - allocations


def get_ocba_X(X, means, vars, delta):
    o = get_ocba(means=means, vars=vars, delta=delta)
    return repeat(X, o, axis=0)



def test_suggest_new_X():
    """
    Test suggest_new_X method
    """
    import numpy as np
    from spotPython.fun.objectivefunctions import analytical
    from spotPython.spot.spot import Spot
    from spotPython.utils.repair import repair_non_numeric
    import pytest

    fun = analytical().fun_sphere

    nn = 3

    spot_test = Spot(
        fun=fun,
        lower=np.array([-10,-1]),
        upper=np.array([10,1]),
        n_points=nn
    )


    # (S-2) Initial Design:
    spot_test.X = spot_test.generate_design(size=spot_test.design_control["init_size"],
                                            repeats=spot_test.design_control["repeats"],
                                            lower=spot_test.lower,
                                            upper=spot_test.upper)
    spot_test.X = repair_non_numeric(spot_test.X,
                                        spot_test.var_type)

    # (S-3): Eval initial design:
    spot_test.y = spot_test.fun(spot_test.X)
    spot_test.surrogate.fit(spot_test.X, spot_test.y)
    X0 = spot_test.suggest_new_X()
    assert X0.shape[0] == nn
    assert X0.shape[1] == 2



def test_extract_from_bounds():
    """
    Test extract_from_bounds
    """
    from spotPython.build.kriging import Kriging
    import numpy as np
    import copy
    import matplotlib.pyplot as plt
    from numpy import append, ndarray, multiply, isinf, linspace, meshgrid, ravel, argmin
    from numpy import ones, zeros, arange, log, var, float64
    from numpy import spacing, empty_like
    from numpy import array
    from spotPython.design.spacefilling import spacefilling

    # One-dim objective function
    ni = 11
    nat_X = linspace(start=0, stop=10, num=ni).reshape(-1, 1)
    nat_y = np.squeeze(nat_X +1)
    S = Kriging(name='kriging', 
                min_theta = -3,
                max_theta = 2,
                seed=124)
    S.nat_X = copy.deepcopy(nat_X)
    S.nat_y = copy.deepcopy(nat_y)
    S.n = S.nat_X.shape[0]
    S.k = S.nat_X.shape[1]
    S.cod_X = empty_like(S.nat_X)
    S.cod_y = empty_like(S.nat_y)
    # assume all variable types are "num" if "num" is
    # specified once:
    if len(S.var_type) == 1:
        S.var_type = S.var_type * S.k    
    S.num_mask = array(list(map(lambda x: x == "num", S.var_type)))
    S.factor_mask = array(list(map(lambda x: x == "factor", S.var_type)))
    S.int_mask = array(list(map(lambda x: x == "int", S.var_type)))
    S.ordered_mask = array(list(map(lambda x: x == "num" or x == "int", S.var_type)))
    S.nat_to_cod_init()
    S.theta = zeros(S.n_theta)
    # TODO: Currently not used:
    S.x0_theta = ones((S.n_theta,)) * S.n / (100 * S.k)
    S.p = ones(S.n_p) * 2.0
    S.pen_val = S.n * log(var(S.nat_y)) + 1e4
    S.negLnLike = None
    S.gen = spacefilling(k=S.k, seed=S.seed)
    # matrix related
    S.LnDetPsi = None
    S.Psi = zeros((S.n, S.n), dtype=float64)
    S.psi = zeros((S.n, 1))
    S.one = ones(S.n)
    S.mu = None
    S.U = None
    S.SigmaSqr = None
    S.Lambda = None
    # build_Psi() and build_U() are called in fun_likelihood
    S.set_de_bounds()
    if S.model_optimizer.__name__ == 'dual_annealing':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds)
    elif S.model_optimizer.__name__ == 'differential_evolution':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds,
                                        maxiter=S.model_fun_evals,
                                        seed=S.seed)
    elif S.model_optimizer.__name__ == 'direct':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds,
                                        # maxfun=S.model_fun_evals,
                                        eps=1e-2)
    elif S.model_optimizer.__name__ == 'shgo':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds)
    elif S.model_optimizer.__name__ == 'basinhopping':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        x0=S.min_X)
    else:
        result = S.model_optimizer(func=S.fun_likelihood, bounds=S.de_bounds)
    # Finally, set new theta and p values and update the surrogate again
    # for new_theta_p_Lambda in de_results["x"]:
    new_theta_p_Lambda = result["x"]
    assert new_theta_p_Lambda.shape[0] == 1

    # 2nd Example:
    from spotPython.build.kriging import Kriging
    import numpy as np
    import spotPython
    from spotPython.fun.objectivefunctions import analytical
    from spotPython.spot import spot
    from spotPython.utils.repair import repair_non_numeric
    fun = analytical().fun_sphere
    lower = np.array([-1,-1])
    upper = np.array([1,1])
    design_control={"init_size": 10}
    surrogate_control={
                "infill_criterion": "y",
                "n_points": 1,
            }
    # Spot: to generate initial design
    S_spot = spot.Spot(fun=fun,
                lower = lower, 
                upper= upper, 
                surrogate=S, 
                fun_evals = 25, 
                noise = False,
                log_level = 50, 
                design_control=design_control, 
                surrogate_control=surrogate_control)

    X = S_spot.generate_design(size=S_spot.design_control["init_size"],
                               repeats=S_spot.design_control["repeats"],
                               lower=S_spot.lower, 
                               upper=S_spot.upper)
    X = repair_non_numeric(X, S_spot.var_type)
    # (S-3): Eval initial design:
    y = fun(X)
    S_spot.min_y = min(y)
    S_spot.min_X = X[argmin(y)]
    # Kriging:
    S = Kriging(name='kriging',
                seed=124, 
                n_theta=2, 
                noise=False, 
                cod_type="norm")
    S.nat_X = copy.deepcopy(X)
    S.nat_y = copy.deepcopy(y)
    S.n = S.nat_X.shape[0]
    S.k = S.nat_X.shape[1]
    S.cod_X = empty_like(S.nat_X)
    S.cod_y = empty_like(S.nat_y)
    # assume all variable types are "num" if "num" is
    # specified once:
    if len(S.var_type) == 1:
        S.var_type = S.var_type * S.k    
    S.num_mask = array(list(map(lambda x: x == "num", S.var_type)))
    S.factor_mask = array(list(map(lambda x: x == "factor", S.var_type)))
    S.int_mask = array(list(map(lambda x: x == "int", S.var_type)))
    S.ordered_mask = array(list(map(lambda x: x == "num" or x == "int", S.var_type)))
    S.nat_to_cod_init()
    S.theta = zeros(S.n_theta)
    # TODO: Currently not used:
    S.x0_theta = ones((S.n_theta,)) * S.n / (100 * S.k)
    S.p = ones(S.n_p) * 2.0
    S.pen_val = S.n * log(var(S.nat_y)) + 1e4
    S.negLnLike = None
    S.gen = spacefilling(k=S.k, seed=S.seed)
    # matrix related
    S.LnDetPsi = None
    S.Psi = zeros((S.n, S.n), dtype=float64)
    S.psi = zeros((S.n, 1))
    S.one = ones(S.n)
    S.mu = None
    S.U = None
    S.SigmaSqr = None
    S.Lambda = None
    # build_Psi() and build_U() are called in fun_likelihood
    S.set_de_bounds()
    if S.model_optimizer.__name__ == 'dual_annealing':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds)
    elif S.model_optimizer.__name__ == 'differential_evolution':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds,
                                        maxiter=S.model_fun_evals,
                                        seed=S.seed)
    elif S.model_optimizer.__name__ == 'direct':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds,
                                        # maxfun=S.model_fun_evals,
                                        eps=1e-2)
    elif S.model_optimizer.__name__ == 'shgo':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds)
    elif S.model_optimizer.__name__ == 'basinhopping':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        x0=S.min_X)
    else:
        result = S.model_optimizer(func=S.fun_likelihood, bounds=S.de_bounds)
    # Finally, set new theta and p values and update the surrogate again
    # for new_theta_p_Lambda in de_results["x"]:
    new_theta_p_Lambda = result["x"]
    assert new_theta_p_Lambda.ndim == 1
    assert new_theta_p_Lambda.shape[0] == 2

    # 3rd example: 2nd example with noise (lambda)
    S = Kriging(name='kriging',
                seed=124, 
                n_theta=2, 
                noise=True, 
                cod_type="norm")
    S.nat_X = copy.deepcopy(X)
    S.nat_y = copy.deepcopy(y)
    S.n = S.nat_X.shape[0]
    S.k = S.nat_X.shape[1]
    S.cod_X = empty_like(S.nat_X)
    S.cod_y = empty_like(S.nat_y)
    # assume all variable types are "num" if "num" is
    # specified once:
    if len(S.var_type) == 1:
        S.var_type = S.var_type * S.k    
    S.num_mask = array(list(map(lambda x: x == "num", S.var_type)))
    S.factor_mask = array(list(map(lambda x: x == "factor", S.var_type)))
    S.int_mask = array(list(map(lambda x: x == "int", S.var_type)))
    S.ordered_mask = array(list(map(lambda x: x == "num" or x == "int", S.var_type)))
    S.nat_to_cod_init()
    S.theta = zeros(S.n_theta)
    # TODO: Currently not used:
    S.x0_theta = ones((S.n_theta,)) * S.n / (100 * S.k)
    S.p = ones(S.n_p) * 2.0
    S.pen_val = S.n * log(var(S.nat_y)) + 1e4
    S.negLnLike = None
    S.gen = spacefilling(k=S.k, seed=S.seed)
    # matrix related
    S.LnDetPsi = None
    S.Psi = zeros((S.n, S.n), dtype=float64)
    S.psi = zeros((S.n, 1))
    S.one = ones(S.n)
    S.mu = None
    S.U = None
    S.SigmaSqr = None
    S.Lambda = None
    # build_Psi() and build_U() are called in fun_likelihood
    S.set_de_bounds()
    if S.model_optimizer.__name__ == 'dual_annealing':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds)
    elif S.model_optimizer.__name__ == 'differential_evolution':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds,
                                        maxiter=S.model_fun_evals,
                                        seed=S.seed)
    elif S.model_optimizer.__name__ == 'direct':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds,
                                        # maxfun=S.model_fun_evals,
                                        eps=1e-2)
    elif S.model_optimizer.__name__ == 'shgo':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        bounds=S.de_bounds)
    elif S.model_optimizer.__name__ == 'basinhopping':
        result = S.model_optimizer(func=S.fun_likelihood,
                                        x0=S.min_X)
    else:
        result = S.model_optimizer(func=S.fun_likelihood, bounds=S.de_bounds)
    # Finally, set new theta and p values and update the surrogate again
    # for new_theta_p_Lambda in de_results["x"]:
    new_theta_p_Lambda = result["x"]
    S.extract_from_bounds(new_theta_p_Lambda)
    assert S.theta.shape[0] == 2
    assert S.Lambda > 0
    assert S.p == np.array([2.])


def test_nat_to_cod_init():
    """
    Test nat_to_cod_init
    """
    from spotPython.build.kriging import Kriging
    import numpy as np
    import copy
    import matplotlib.pyplot as plt
    from numpy import append, ndarray, multiply, isinf, linspace, meshgrid, ravel
    from numpy import ones, zeros, arange
    from numpy import spacing, empty_like
    from numpy import array

    # One-dim objective function
    ni = 11
    nat_X = linspace(start=0, stop=10, num=ni).reshape(-1, 1)
    nat_y = np.squeeze(nat_X +1)
    S = Kriging(name='kriging',  seed=124)
    S.nat_X = copy.deepcopy(nat_X)
    S.nat_y = copy.deepcopy(nat_y)
    S.n = S.nat_X.shape[0]
    S.k = S.nat_X.shape[1]
    S.cod_X = empty_like(S.nat_X)
    S.cod_y = empty_like(S.nat_y)
    # assume all variable types are "num" if "num" is
    # specified once:
    if len(S.var_type) == 1:
        S.var_type = S.var_type * S.k
    S.num_mask = array(list(map(lambda x: x == "num", S.var_type)))
    S.factor_mask = array(list(map(lambda x: x == "factor", S.var_type)))

    S.nat_to_cod_init()
    assert S.nat_X.ndim == 2
    assert S.nat_X.shape[0] == ni
    assert S.nat_y.ndim == 1
    assert S.nat_y.shape[0] == ni
    #
    assert S.cod_X.ndim == 2
    assert S.cod_X.shape[0] == ni
    assert S.cod_y.ndim == 1
    assert S.cod_y.shape[0] == ni